import inspect
import logging
from logging import Logger
from typing import (
    Type,
    TypeVar,
    Any,
    Optional,
    Dict,
    Set,
    Union,
    get_args
)

from ._DependencyContext import DependencyContext
from ._DependencyMap import DependencyMap
from ._utils import (
    get_class_parent_types
)
from .exception import DependencyNotFoundException, AbstractDependencyException

AnyType = TypeVar("AnyType")


class DependencyContextImpl(DependencyContext):

    def __init__(self, dependency_map: DependencyMap, logger: Optional[Logger] = None):
        self._dependency_map = dependency_map

        self.dependency_instances = {}
        self.dependency_extra_parameters: Dict[AnyType, Dict[str, Any]] = {}
        self.logger = logger or logging.getLogger()
        self._dispose_order = []
        self.add_dependency_instance(self)
        self.add_dependency_instance(self.logger)

    def add_dependency_instance(self, dependency_instance: Any):
        self.dependency_instances[type(dependency_instance)] = dependency_instance
        for base_class in get_class_parent_types(type(dependency_instance)):
            self.dependency_instances[base_class] = dependency_instance

    def add_dependency_extra_parameter(
        self,
        parameter_name: str,
        parameter_value: Any,
        for_type: Optional[AnyType] = None
    ):
        if for_type not in self.dependency_extra_parameters:
            self.dependency_extra_parameters[for_type] = {}
        self.dependency_extra_parameters[for_type][parameter_name] = parameter_value

    def get_instance(
        self,
        cls: Type[AnyType],
        params: Optional[Dict[str, Any]] = None
    ) -> AnyType:
        return self._get_instance(
            cls,
            params=params,
            ignore_in_map=True,
            is_optional=False
        )

    def _get_instance(
        self,
        cls: Type[AnyType],
        params: Optional[Dict[str, Any]] = None,
        ignore_in_map: Optional[bool] = False,
        is_optional: Optional[bool] = False,
    ) -> AnyType:
        if cls is None:
            return None

        try:
            if isinstance(self, cls):
                return self
        except TypeError:
            pass

        if cls in self.dependency_instances:
            return self.dependency_instances[cls]

        if cls in self._dependency_map.builders:
            builder_instance = self._create_instance(
                cls=self._dependency_map.builders[cls],
                params=params
            )
            dependency_instance = builder_instance.build()
            dependency_type = type(dependency_instance)
            self.logger.debug(f"Built instance of {dependency_type.__name__}")
            self.add_dependency_instance(dependency_instance)
            if dependency_type not in self._dependency_map.builders:
                self._dependency_map.builders[dependency_type] = self._dependency_map.builders[cls]
                for child_type in get_class_parent_types(dependency_type):
                    self._dependency_map.builders[child_type] = self._dependency_map.builders[cls]

            self._dispose_order.append(dependency_instance)
            self._dispose_order.append(builder_instance)

            return dependency_instance

        dependency_class = None
        if ignore_in_map and not inspect.isabstract(cls):
            dependency_class = cls
        elif cls in self._dependency_map.classes:
            dependency_class = self._dependency_map.classes[cls]

        if dependency_class is not None:
            dependency_instance = self._create_instance(
                cls=dependency_class,
                params=params
            )
            if dependency_instance is not None:
                self._dispose_order.append(dependency_instance)
                return dependency_instance

        if cls in self._dependency_map.classes:
            dependency_class = self._dependency_map.classes[cls]
            dependency_instance = self._create_instance(
                cls=dependency_class,
                params=params
            )
            self._dispose_order.append(dependency_instance)
            return dependency_instance

        if not is_optional:
            raise DependencyNotFoundException(cls)

        return None

    def get_dependencies_by_parent(self, parent_type: Type[AnyType]) -> Set[Type[AnyType]]:
        return self._dependency_map.classes_by_parent.get(parent_type, set())

    def has_instance(self, cls: Type):
        return cls in self.dependency_instances

    def _create_instance(
        self,
        cls: Type,
        params: Optional[Dict[str, Any]] = None,
        only_create: Optional[bool] = False
    ):
        if cls in self.dependency_instances:
            return self.dependency_instances[cls]

        if inspect.isabstract(cls):
            raise AbstractDependencyException(cls)

        self.logger.debug(f"Creating instance of {cls.__name__}")
        signature = inspect.signature(cls.__init__)
        factored_constructor_args = {}
        for parameter in signature.parameters.values():
            if parameter.name in ["self", "args", "kwargs"]:
                continue

            if params is not None and parameter.name in params:
                factored_constructor_args[parameter.name] = params[parameter.name]
            elif cls in self.dependency_extra_parameters and parameter.name in self.dependency_extra_parameters[cls]:
                factored_constructor_args[parameter.name] = self.dependency_extra_parameters[cls][parameter.name]
            elif None in self.dependency_extra_parameters and parameter.name in self.dependency_extra_parameters[None]:
                factored_constructor_args[parameter.name] = self.dependency_extra_parameters[None][parameter.name]
            else:
                build_cls = parameter.annotation
                # Validation for parameters Optional[some_type]
                if getattr(build_cls, "__origin__", None) == Union:
                    args = get_args(build_cls)
                    if len(args) > 0:
                        build_cls = get_args(build_cls)[0]
                    else:
                        build_cls = None

                try:
                    factored_constructor_args[parameter.name] = self._get_instance(
                        cls=build_cls,
                        is_optional=not (parameter.default is parameter.empty)
                    )
                except DependencyNotFoundException as ex:
                    raise DependencyNotFoundException(ex.cls, f"{ex.path} -> {ex.cls}.{parameter.name}")

        dependency_instance = cls(**factored_constructor_args)
        self._check_created_method(dependency_instance)

        if not only_create:
            self.add_dependency_instance(dependency_instance)

        return dependency_instance

    def dispose(self):
        destroyed_instances = set()
        for instance in self._dispose_order:
            if instance not in destroyed_instances:
                instance_type = type(instance)
                if instance_type in self._dependency_map.builders:
                    builder_type = self._dependency_map.builders[instance_type]
                    builder_instance = self.dependency_instances.get(builder_type, None)
                    if builder_instance is not None:
                        builder_instance.destroy(instance)
                        destroyed_instances.add(instance)
                elif self._check_destroy_method(instance):
                    destroyed_instances.add(instance)

    def _check_created_method(self, dependency_instance):
        if hasattr(dependency_instance, '__enter__') and callable(getattr(dependency_instance, '__enter__')):
            self.logger.debug(f"Created instance of {type(dependency_instance).__name__}")
            dependency_instance.__enter__()
            return True
        return False

    def _check_destroy_method(self, instance):
        if hasattr(instance, '__exit__') and callable(getattr(instance, '__exit__')):
            self.logger.debug(f"Destroying instance of {type(instance).__name__}")
            instance.__exit__(exc_type=None, exc_val=None, exc_tb=None)
            return True
        return False



from typing import Type


class DependencyNotFoundException(Exception):
    def __init__(self, cls: Type, path: str = ""):
        self.cls = cls
        self.path = path
        super().__init__(f"Dependency {cls.__name__} not found while looking for {path}")



from typing import Dict, Type, Set

from dependencyinjection import DependencyBuilder


class DependencyMap:
    def __init__(
        self,
        classes: Dict[Type, Type] = None,
        classes_by_parent: Dict[Type, Set[Type]] = None,
        builders: Dict[Type, Type[DependencyBuilder]] = None,
    ):
        self.classes = classes or {}
        self.classes_by_parent = classes_by_parent or {}
        self.builders = builders or {}

    def clone(self):
        return DependencyMap(
            classes={**self.classes},
            classes_by_parent={**self.classes_by_parent},
            builders={**self.builders},
        )



import importlib
import inspect
import os
import pkgutil
import sys
from abc import ABC
from logging import Logger, getLogger
from types import ModuleType
from typing import Union, Type, Callable, Any, Optional


def get_method_by_name(element: Union[Type, object], name: str):
    for method_name, method_obj in inspect.getmembers(element):
        if (
                (inspect.isfunction(method_obj) or inspect.ismethod(method_obj))
                and method_name == name
        ):
            return method_obj
    return None


def inherit_from(element: Union[Type, object], parent_type: Type):
    return (
            isinstance(element, parent_type)
            or (isinstance(element, type) and parent_type in element.__bases__)
    )


def load_module_then_execute(
        load_module: ModuleType,
        callback: Callable[[ModuleType, ], Any],
        logger: Optional[Logger] = None,
        recursive: Optional[bool] = True
):
    if recursive:
        logger = logger or getLogger()
        package_dir = os.path.dirname(load_module.__file__)

        for (_, module_name, _) in pkgutil.iter_modules([package_dir]):
            if os.path.exists(os.path.join(package_dir, module_name, '__init__.py')):
                full_package_name = f"{load_module.__name__}.{module_name}"

                if full_package_name not in sys.modules:
                    logger.debug(f"Load extra module {full_package_name}")
                    module = importlib.import_module(full_package_name)
                else:
                    logger.debug(f"Using loaded module {full_package_name}")
                    module = sys.modules[full_package_name]

                load_module_then_execute(module, callback, logger)

    callback(load_module)


def get_class_parent_types(dep_class: Type):
    return [child for child in dep_class.__bases__ if child not in [ABC, object]]



from abc import abstractmethod, ABC
from typing import Generic, TypeVar

AnyType = TypeVar("AnyType")


class DependencyBuilder(Generic[AnyType], ABC):

    @abstractmethod
    def build(self) -> AnyType:
        raise NotImplementedError()

    @abstractmethod
    def destroy(self, instance: AnyType) -> None:
        raise NotImplementedError()



import inspect
import logging
from logging import Logger
from types import ModuleType
from typing import List, Type, Union, Optional, Callable, Set

from ._DependencyBuilder import DependencyBuilder
from ._DependencyContext import DependencyContext
from ._DependencyContextImpl import DependencyContextImpl
from ._DependencyMap import DependencyMap
from ._utils import (
    load_module_then_execute,
    get_class_parent_types,
    get_method_by_name,
    inherit_from,
)


class DependencyLoader(object):

    def __init__(self, logger: Optional[Logger] = None):
        self.dependency_map = DependencyMap()
        self.logger = logger or logging.getLogger()

    def load_modules(
            self,
            modules: List[ModuleType],
            recursive: Optional[bool] = True,
            class_filter: Optional[Callable[[Type, ], bool]] = None,
    ):
        for module in modules:
            load_module_then_execute(
                load_module=module,
                callback=lambda m: self._load_dependencies_from_module(
                    load_module=m,
                    class_filter=class_filter
                ),
                logger=self.logger,
                recursive=recursive
            )

    def build_context(self) -> DependencyContext:
        return DependencyContextImpl(self.dependency_map.clone(), self.logger)

    def get_dependencies_by_parent(self, parent_type: Type) -> Set[Type]:
        return self.dependency_map.classes_by_parent.get(parent_type, [])

    def add_dependency_class(self, dep_class: Type):
        if not inspect.isabstract(dep_class):
            self.logger.debug(f"---> Loaded class dependency {dep_class.__name__}")
            if self._is_builder(dep_class):
                build_method = inspect.signature(self._get_builder_method(dep_class))
                build_return = build_method.return_annotation
                self.dependency_map.builders[build_return] = dep_class
                for child_type in get_class_parent_types(build_return):
                    self.dependency_map.builders[child_type] = dep_class
            else:
                self._add_class_hierarchy_to_dependency(dep_class)

            for base_type in get_class_parent_types(dep_class):
                if base_type not in self.dependency_map.classes_by_parent:
                    self.dependency_map.classes_by_parent[base_type] = set()
                self.dependency_map.classes_by_parent[base_type].add(dep_class)

    def _add_class_hierarchy_to_dependency(self, cls: Type):
        if cls not in self.dependency_map.classes:
            self.dependency_map.classes[cls] = cls
            for parent in get_class_parent_types(cls):
                self.dependency_map.classes[parent] = cls

    def _load_dependencies_from_module(
            self,
            load_module: ModuleType,
            class_filter: Optional[Callable[[Type, ], bool]] = None,
    ):
        for member_name, member_obj in inspect.getmembers(load_module):
            if inspect.isclass(member_obj) and not inspect.isabstract(member_obj):
                if class_filter is not None and class_filter(member_obj):
                    self.add_dependency_class(member_obj)
                elif class_filter is None:
                    self.add_dependency_class(member_obj)

    @staticmethod
    def _get_builder_method(element: Union[Type, object]):
        return get_method_by_name(element, 'build')

    @staticmethod
    def _is_builder(dep_class: Union[Type, object]):
        return inherit_from(dep_class, DependencyBuilder)



from ._DependencyBuilder import AnyType
from ._DependencyBuilder import DependencyBuilder
from ._DependencyContext import DependencyContext
from ._DependencyLoader import DependencyLoader
from ._DependencyMap import DependencyMap



from abc import ABC, abstractmethod
from typing import Any, Type, Optional, Set, Dict

from . import AnyType


class DependencyContext(ABC):

    @abstractmethod
    def add_dependency_instance(self, dependency_instance: Any):
        raise NotImplementedError()

    @abstractmethod
    def add_dependency_extra_parameter(
        self,
        parameter_name: str,
        parameter_value: Any,
        for_type: Optional[AnyType] = None
    ):
        raise NotImplementedError()

    @abstractmethod
    def get_instance(self, cls: Type[AnyType], params: Optional[Dict[str, Any]] = None) -> AnyType:
        raise NotImplementedError()

    @abstractmethod
    def get_dependencies_by_parent(self, parent_type: Type[AnyType]) -> Set[Type[AnyType]]:
        raise NotImplementedError()

    @abstractmethod
    def has_instance(self, cls: Type):
        raise NotImplementedError()

    @abstractmethod
    def dispose(self):
        raise NotImplementedError()
